import json
import os
from functools import partial

from dateutil.parser import parse as parse_date

import gevent


import mechanize
import requests
import websocket

INFINITY = float('inf')

class Client(object):
  def __init__(self, host, auth_input, port=80, cookie_path="cookie3.txt"):
    self.host = host
    self.port = port
    self.auth_input = auth_input
    self.cookie_path = cookie_path
        
    self.br = mechanize.Browser()
    self.br.set_handle_robots(False)
    
    cj = self.cookiejar = mechanize.LWPCookieJar()
    if os.path.exists(cookie_path):
      cj.load(cookie_path)

    self.br.set_cookiejar(cj)
    self.login()
    assert self.br.geturl().endswith('integrated')
    cj.save(cookie_path)


  @property
  def base_url(self):
    return "http://{host}:{port}".format(
      host = self.host,
      port = self.port
    )

  def get(self, relative_url, *args, **kw):
    return requests.get(self.base_url + relative_url, cookies=self.cookies, *args, **kw)
    
  def post(self, relative_url, data=None):
    return requests.post(self.base_url + relative_url, data, cookies=self.cookies)

    
  @property
  def cookies(self):
    cookies = {}
    for cookie in self.cookiejar:
      if cookie.domain == self.host:
        cookies[cookie.name] = cookie.value
    return cookies

  def projects(self):
    resp = self.get("/workspaces/")
    resp.raise_for_status()
    projects = resp.json()

    return map(lambda d: Project(session=self, **d), projects)
    
  def login(self):
    br = self.br

    br.open(self.base_url)

    if not br.geturl().endswith('integrated'):
      # login with github
      response = br.follow_link()

      if br.geturl().endswith('integrated'):
        # github redirected us back, we're good to go
        return
      else:
        br.select_form(nr=0)
        br['login'], br['password']  = self.auth_input()
        br.submit()
      
  def create_project(self, **kw):
    resp = self.post("/workspaces", kw)
    return resp.content
    
  def rebuild(self, project_id, keep):
    if keep:
      keep = '?keep=true'
    else:
      keep = ''
    self.post("/workspaces/{}/rebuild{}".format(project_id, keep))
    
  def rebuild_segment(self, project_id, table, segment):
    self.post("/workspaces/{}/{}/{}/rebuild".format(project_id, table, segment))



class Project(object):
  def __init__(self, **kw):
    self.set_attrs(kw)
    
  def set_attrs(self, dict):
    for key,value in dict.items():
      setattr(self, key, value)
      
  def __repr__(self):
    return "Project: {}  queued: {} failed: {} built: {}".format(
      self.title,
      len(self.queued),
      len(self.failed),
      len(self.built),
      
    )

  @property
  def segments(self):
    collection = []

    for table in self.tables.values():
      collection.extend(table['segments'])
    return collection

  def filter(self, status):
    return [s for s in self.segments if s['status'] == status]
  
  @property
  def built(self):
    return self.filter('built')
  
  @property
  def building(self):
    # todo: have server return building
    return []
    
  @property
  def queued(self):
    return self.filter('pending')
  
  @property  
  def failed(self):
    return self.filter('failed')


  def next_build_in(self):
    # returns the number of seconds until the next build or INFINITY
    # if the project well never be built (errors, paused, etc)

    if hasattr(self, 'next_build_at') and self.next_build_at is not None:
      interval = parse_date(self.next_build_at) - parse_date(self.current_date_time)
      return interval.total_seconds()
    else:
      return INFINITY
  
  def query(self, query, **headers):
    return self.session.get(
      '/workspaces/{}/query'.format(self.id), 
      params={'q': query},
      headers=headers,
      stream=True
    )
    
    
  def resume(self):
    return self.session.post(
      '/workspaces/{}/resume'.format(self.id)
    )
    

  def simulate(self, target, **headers):
    return self.session.get(
      '/workspaces/{}/simulate/{}'.format(self.id, target),
      headers=headers,
      stream=True
    )
    
  def step(self, target):
    return self.session.post(
      '/workspaces/{}/step/{}'.format(self.id, target),
    )
  
    
        
  def import_repo(self, url, owner, name, branch='master'):

    return self.session.post(
      '/workspaces/{}/import_repo'.format(self.id),
      {
        'url':    url,
        'owner':  owner,
        'name':   name,
        'branch': branch
      }
    )
    


  @property
  def repositories(self):
    return self.session.get('/workspaces/{}/repos'.format(self.id)).json()
      
  def subscribe(self, callback):

    def on_open(socket, *args, **kw):
      gevent.spawn(ping, socket)
      socket.send("1::/"+self.id)

    def on_message(socket, msg, *args, **kw):
      code, data, namespace, payload = msg.split(':',3)
      code = int(code)
      if code == 1:
        gevent.spawn(callback,self, {})
      elif code == 5:
        event = json.loads(payload)
        gevent.spawn(callback, self, *event['args'])

    def on_error(socket, msg, *args, **kw):
      print 'error', msg, args, kw
      
    hand_shake_url ='/socket.io/1/'
    #resp = requests.get(hand_shake_url, cookies=c)

    resp = self.session.post(hand_shake_url)

    sid,heartbeat_timeout,connection_timeout,accetped = resp.content.split(':')
    sid,heartbeat_timeout,connection_timeout = map(int, (
      sid,heartbeat_timeout,connection_timeout
    ))
    
    ws_url = 'ws://{}:{}/socket.io/1/websocket/{}'.format(
      self.session.host,
      self.session.port,
      sid
    )
    
    def ping(socket):
      timeout = max(heartbeat_timeout - 1, 1)
      while socket.sock.connected:
        socket.send("2::")
        gevent.sleep(timeout)
    
    
    headers = ['Cookie: %s' % resp.request.headers['Cookie']]
    #from gevent import monkey; monkey.patch_all()
    ws = websocket.WebSocketApp(
      ws_url,
      header=headers,
      on_open   =on_open,
      on_message = on_message,
      on_error = on_error
    )

    ws.run_forever()
  
  



import os.path
from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))

README_PATH = os.path.join(HERE, 'README.rst')
try:
    README = open(README_PATH).read()
except IOError:
    README = ''


try:
  REQUIRES=open(os.path.join(HERE, "requirements.txt")).readlines()
except IOError:
  REQUIRES=[]


setup(
  name='trivio',
  packages=find_packages(),
  version='0.2.0c',
  description='triv.io command line client',
  long_description=README,
  author='Scott Robertson',
  author_email='scott@triv.io',
  url='http://github.com/trivio/trivio.client',
  classifiers=[
      "Programming Language :: Python",
      "License :: OSI Approved :: MIT License",
      "Operating System :: OS Independent",
      "Development Status :: 3 - Alpha",
      "Environment :: Web Environment",
      "Intended Audience :: Developers",
      "Topic :: Software Development",
  ],
  entry_points = {
    'console_scripts': [
      'trivio = triv.io.client:main',
    ]
  },
  install_requires=REQUIRES,
)



#!/usr/bin/env python

import json
import os
import sys
import re
import time
from urlparse import urlparse

import api

import gevent


INFINITY = float('inf')


PROJECT_FILE='.trivio.project'

COOKIE_JAR_PATH="cookie3.txt"
PARSE_GITHUB_URL=re.compile('git@github.com:(\w+)/(\w+).git')


default_url="http://app.triv.io:80"

def auth_input():
  login = raw_input('login: ')
  password = raw_input('password: ')
  return login, password

def format_delay(seconds):
  if seconds == INFINITY:
    return "forever"
  
  seconds = int(seconds)
  minutes = 60
  hours   = minutes * 60
  days    = hours * 24
  
  t = []
  for label, part in [('d', days), ('h',hours), ('m',minutes)]:
    whole, seconds = divmod(seconds, part)
    if whole == 0 and len(t) == 0:
      continue
    else:
      t.append("{:02d}{}".format(whole,label))
  t.append("{:02d}{}".format(seconds,'s'))

    
  return " ".join(t)

def rules_view(project):
  row_format = "{:40} {:>8} {:>8} {:>8} {:>8}"
  
  # generates one empty string per field applies it to the
  # format and measures the length of the string
  row_length = len(row_format.format(*(['']*5)))
  
  clearline()
  print row_format.format("", "Queued", "Building","Failed", "Built")
  
  print '-' * row_length
  
  print row_format.format(
    "All", 
    len(project.queued),
    len(project.building),
    len(project.failed),
    len(project.built)
  )
  line_count = 3

  for title, table in project.tables.items():
    #title = rule['target']
    #table = project.tables[title]
    
    if(len(title) > 40):
      title = '...' + title[-37:]
      
    counts = {}
    for segment in table['segments']:
      status = segment['status']
      counts[status] = counts.get(status,0) + 1
  
    print row_format.format(
      title,
      counts.get('queued', 0),
      counts.get('building', 0),
      counts.get('failed', 0),
      counts.get('built', 0)
    )
    line_count += 1
  
  return line_count
    
  

def clearline(s=''):
  sys.stdout.write("\r\x1b[K" + s)
  sys.stdout.flush()

def moveup(n=1):
  sys.stdout.write("\r\033[{:d}A".format(n))
  sys.stdout.flush()


def current_project(session):
  """
  
  Returns the current project or None if the trivio command was run from
  outside a trivio project directory..
  
  This method searches the current directory for a file named ".trivio.project"
  if it can't be found it will search each ancestor directory.
  """
  path = os.getcwd()
  
  while path != '/':
    try:
      project_props = json.load(open(os.path.join(path, PROJECT_FILE)))
      print "Changing to dir " + path
      os.chdir(path)
      return api.Project(session=session, **project_props)
    except IOError:
      path = os.path.dirname(path)

def current_target(settings_path):
  settings = None

  if os.path.exists(settings_path):
    content = open(settings_path).read()
    if content:
      settings = json.loads(content)
    else:
      settings = None

  if settings is None:
    settings = target_cmd(settings_path)
    
  return settings['target_url']

def create_cmd(client, title, *repos):
  """
  Creates a new project, ensuring the user is not under a project directory first.
  """
  assert current_project(client) is None
  
  print "Creating project directory"
  os.mkdir(title)
  print "Creating repositories"
  os.mkdir(os.path.join(title, "repositories"))
  
  
  print "Creating trivio project"
  project = client.create_project(title=title)
  open(os.path.join(title, PROJECT_FILE), 'w').write(project)


def checkout_cmd(client, title_or_id=None):
  assert current_project(client) is None
  if title_or_id is None:
    available=client.projects()
  else:

    available = []
    for project in client.projects():
      if (project.title and project.title.startswith(title_or_id))  or project.id.startswith(title_or_id):
        available.append(project)
        
  if len(available) == 1:
    project = available[0]
  else:
    # to many projects match, prompt the user
    print "Which project would you like to work on?"
    for i, project in enumerate(available):
      print i,' ', project
    
    index = int(raw_input("Enter number of projct to workon: "))
    project = available[index]


  if os.path.exists(project.title):
    print "Project {} already exists".format(project.title)
  else:    
    os.mkdir(project.title)
    os.chdir(project.title)
    open(PROJECT_FILE, 'w').write(json.dumps(
      dict(
        id=project.id,
        title=project.title
      )
    ))
    
    os.mkdir('repositories')
    os.chdir('repositories')
    return pull_cmd(client)

def import_cmd(client, git_url):
  project = current_project(client)
  
  owner, repo = PARSE_GITHUB_URL.search(git_url).groups()
  
  repo = repo.split('.git')[0]

  resp = project.import_repo(git_url, owner, repo)
  pull_cmd(client)

def remove_repo_cmd(client, path):
  project = current_project(client)
  print path

def errors_cmd(client, project_id=None):
  project = current_project(client)
  print project.rule_errors, project.rules
  for error in project.rule_errors:
    print error
    
def remove_cmd(client, project_id=None):
  """
  Removes the idetified project, if no project was specified prompts the user
  to select the project to remove.
  """
  if project_id is None:
    projects = client.projects()

    for i, p in enumerate(projects):
      print "  {}: {.title}".format(i,p)
    x = int(raw_input("Enter project to remove: "))
    project_id = projects[i].id
  print "removing " + project_id
    
def target_cmd(settings_path, target_url=None):
  if target_url is None:
    target_url = raw_input("Enter URL [{}]:".format(default_url))
    if target_url.strip() == '':
      target_url = default_url
      
  settings = dict(target_url=target_url)
  open(settings_path, 'w').write(json.dumps(settings))
  return settings
 
def list_cmd(client):
  """Print a list of each project for the logged in user."""

  for project in client.projects():
    print project
    
    
def pull_cmd(client):
  """Updates the repos in the project directory to match the project on trivio"""
  
  project = current_project(client)
  os.chdir('repositories')
  
  for repo in  project.repositories:
    if not os.path.exists(repo['owner']):
      os.mkdir(repo['owner'])
    os.chdir(repo['owner'])
    if not os.path.exists(repo['name']):
      print "Cloning " + repo['name']
      os.system('git clone ' + repo['git_url'])
    else:
      print "Updating " + repo['name']      
      os.chdir(repo['name'])
      os.system('git pull')
      os.chdir('..')
    os.chdir('..')
      

 
def push_cmd(client, msg="pushing", keep=False):
  
  """
  Rebuilds the project after commiting and pushing changes for any modified
  repository.
  """

  if isinstance(keep, basestring):
    keep = keep.lower().strip()
    if keep == 'true':
      keep = True
    elif keep == 'false':
      keep = False
    else:
      raise ValueError('invalid value "{}"'.format(keep))

  project = current_project(client)
  # getting the current project should put us in the correct directory  

  curdir = os.getcwd()
  os.chdir('repositories')
  for owner in os.listdir('.'):
    print "changing to " + owner
    os.chdir(owner)
    for repo in os.listdir('.'):
      os.chdir(repo)
      os.system('git commit -am "{}"; git push'.format(msg))
      os.chdir('..')
    os.chdir('..')
  client.rebuild(project.id, keep)

def rebuild_cmd(client, table=None, segment=None):
  project = current_project(client)
  if (table, segment) == (None, None):
    client.rebuild(project.id, keep=False) 
  else:
    client.rebuild_segment(project.id, table, segment) 

    
   
def query_cmd(client, query_str, accept="application/x-json-stream"):
  print query_str
  project = current_project(client)

  for chunk in project.query(query_str, accept=accept).iter_content(chunk_size=10240):
    sys.stdout.write(chunk)

def resume_cmd(client):
  project = current_project(client)
  print project.resume().content
  

def watch_cmd(client):
  project = current_project(client)
  from gevent import monkey; monkey.patch_all(thread=False)
  gevent.spawn(project.subscribe,on_project_update).join()
  
def on_project_update(project, attrs):
  project.set_attrs(attrs)
  print project
  line_count = rules_view(project)
  
  seconds = project.next_build_in()
  while seconds > 0:
    time.sleep(1)
    seconds -= 1

    clearline(
      "Next build in {} seconds".format(format_delay(seconds))
    )
  print
  
  
  
def simulate_cmd(client, target, accept='application/x-json-stream'):
  project = current_project(client)
  for chunk in project.simulate(target, accept=accept).iter_content(chunk_size=10240):
    sys.stdout.write(chunk)

def step_cmd(client, target):
  project = current_project(client)
  print project.step(target).content


def main():
  args = sys.argv[1:]
  command = args[0]
  
  # setup cookie dir
  trivio_dir = os.path.expanduser('~/.triv.io')
  if not os.path.exists(trivio_dir):
    os.mkdir(trivio_dir)

  settings_path = os.path.join(trivio_dir, 'settings.json')

  if command == 'target':
    target_cmd(settings_path)
    sys.exit(0)
  else:
    target_url =  current_target(settings_path)
  
  url = urlparse(target_url)
  
  client = api.Client(
    url.hostname,
    auth_input,
    port=url.port,
    cookie_path = os.path.join(trivio_dir, 'cookies.txt')
  )    
  
  method = globals().get(command+'_cmd', None)
  if method is None:
    method = query_cmd
  else:
    args.pop(0)
    

  try:
    try:
      method(client, *args)
    except KeyboardInterrupt:
      pass
  finally:
    print
    


  
if __name__ == "__main__":
  main()
